import numpy as np
from numpy import cos,sin,pi
import os

from fc_tools.others import func2str,isModuleFound,is_function
from fc_oogmsh import gmsh
from fc_simesh.siMesh import siMesh
from fc_vfemp1.operators import Loperator,Lmass
from fc_vfemp1.BVP import BVP,PDE
from fc_vfemp1.FEM import AssemblyP1 

from fc_vfemp1_surface.sys import get_geo
import fc_vfemp1_surface.lib as slib
from fc_vfemp1_surface.examples import sBVPsamples


def run(**kwargs):
  N=kwargs.pop('N',30)
  verbose=kwargs.pop('verbose',1)
  u=kwargs.pop('u',sBVPsamples.uxyz01)
  f=kwargs.pop('f',sBVPsamples.fxyz01)
  C=kwargs.pop('C',None)
  Fmap=kwargs.pop('Fmap',sBVPsamples.Fmapxyz01)
  graphics=kwargs.pop('graphics',1)
  geofile='sphere4surf'
  (geodir,geofile)=get_geo(3,2,geofile)
  geoFile=geodir+os.sep+geofile+'.geo'
  if verbose>0:
    slib.print_separator()
    slib.PrintCopyright()
    slib.print_separator()
    print('      Surface BVP on unit sphere maps by function Fmap')
    print('      Mesh generated with gmsh and %s.geo file.\n'%geofile)
    print('               -laplacian(u) = f\n')
    print(' with ')
    print('   Fmap = %s'%func2str(Fmap,source=False))
    print('   u = C + %s'%func2str(u,source=False))
    if C is not None:
      print('   C = %g'%C)
    print('   f = %s'%func2str(f,source=False))
    slib.print_separator()
    if verbose==10:
      return
  print('*** Building the mesh using GMSH')
  meshfile=gmsh.buildmesh3ds(geodir+os.sep+geofile+'.geo',N, force=False,verbose=0)
  Th=siMesh(meshfile,trans=Fmap)
  print('     Mesh sizes : nq=%d, nme=%d, h=%.3e'%(Th.nq,Th.get_nme(),Th.get_h()))
  print('*** Setting the surface BVP')
  Lop=Loperator(dim=3,d=2,A=[[1,None,None],[None,1,None],[None,None,1]])
  pde=PDE(Op=Lop,f=f)
  bvp=BVP(Th,pde=pde)
  print('*** Solving the surface BVP')
  U=bvp.solve(condition=[0])
  if C is None:
    Mass=AssemblyP1(Th,Lmass(Th.dim))
    C=-np.sum(Mass*Th.feval(u))/(4*np.pi)
  uex=lambda x,y,z: u(x,y,z)+C
  
  if graphics and isModuleFound('fc_simesh_mayavi'):
    if verbose>0:
      print('*** Plotting results')
    from mayavi import mlab
    import fc_simesh_mayavi.siMesh as simlab
    from fc_vfemp1_surface.mayavi import plotSBVPsolution
    plotSBVPsolution(Th,U,uex=uex,**kwargs)
  
  return {'bvp':bvp,'N':N,'uex':uex,'f':f,'Fmap':Fmap,'U':U}
